#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>


#ifndef F_CPU
	#define F_CPU 8000000
#endif

	#define UART_BAUD_RATE 31250

#define RX_BUF_SIZE 32
#define TX_BUF_SIZE 32
unsigned char rx_buf[RX_BUF_SIZE], tx_buf[TX_BUF_SIZE], rx_rdp=0, rx_wrp=0, tx_rdp=0, tx_wrp=0;


// MCU initialisieren
void hw_init(void) {
	// Ports
	DDRB = 0x3F;	// Output-Port fr Relais, Pins 0..5 werden verwendet
	PORTB = 0x3F;	// Ausgang ist auf 5V

	DDRD = 0x02;	// Input-Port, auer TxD
	PORTD = 0xFF;	// Pull-Ups aktivieren

	// Timer
	TCCR0 = 0;
	TCNT0 = 0;

	TCCR1A = 0;
	TCCR1B = 0;//x04;	// Vorteiler 256 -> 31250kHz
	TCNT1 = 0;

	// Interrupts
	// mega8
	TIMSK = 0x51;	// Compare Match A Int. on, Overflow Timer 0 on
	TIFR = 0x51;	// Flag lschen

	// UART
	// Hilfsmakro zur UBRR-Berechnung ("Formel" laut Datenblatt)
	#define UART_UBRR_CALC(BAUD_,FREQ_) ((FREQ_)/((BAUD_)*16L)-1)
	UBRRH = (uint8_t)(UART_UBRR_CALC(UART_BAUD_RATE,F_CPU)>>8);
	UBRRL = (uint8_t)UART_UBRR_CALC(UART_BAUD_RATE,F_CPU);
	UCSRB = (1<<RXEN)|(1<<TXEN);//|(1<<RXCIE);//|(1<<UDRIE);
	UDR = 'A';

//	sei();
}


void send_byte(void) {
	if (UCSRA & (1<<UDRE)) {
		UDR = tx_buf[tx_rdp++];
		tx_rdp &= TX_BUF_SIZE-1;
	}
}


unsigned char getdata(void) {
	unsigned char b;
	do{
		while (!(UCSRA & (1<<RXC)));
		b = UDR;
	}while (b == 0xf8 || b == 0xfe);
	return b;
}


void sendbyte(unsigned char byte) {
	while (!(UCSRA & (1<<UDRE)));
	UDR = byte;
}


void main(void) {
	unsigned char cnt=0, ev, dp[4], adp, len;
	hw_init();

/*	sendbyte(0x90);
	sendbyte(0x50);
	sendbyte(0x70);*/
/*	while (1) {
		ev = getdata();
		PORTB = ~ev;
	}
*/
	while (1) {
		ev = getdata();
		

		// Running status
		if (ev < 0x80) {
			dp[0] = ev;
			ev = adp;
			cnt = 1;
			if ((ev&0xf0)==0x90 || (ev&0xf0)==0x80)
				UDR = 127-dp[0];
			else
				UDR = dp[0];
		}
		else {
			cnt = 0;
			adp = ev;
			UDR = ev;
		}
		//PORTB = ~ev;

		// MidiEvent
		switch (ev) {
		case 0xf0:	// SysEx
			do {
				ev = getdata();
				UDR = ev;
			} while (ev != 0xf7);
			ev = 0xf0;
			break;
		case 0xf2:
			len = 2;
			break;
		case 0xf3:
			len = 1;
			break;
		default:
			switch (ev & 0xf0) {
			case 0xf0:
				len = 0;
				break;
			case 0xc0:	// ProgramChange
			case 0xd0:	// ChannelAftertouch
				len = 1;
				break;
			case 0x80:	// NoteOff
			case 0x90:	// NoteOn
			case 0xa0:	// PolyphonicAftertouch
			case 0xb0:	// ControlModeChange
			case 0xe0:	// PitchWheel
				len = 2;
				break;
			default:
				len = 0;
				PORTB = 0;
			}
		}
		if ((ev&0xf0) != 0xf0) {
			while (cnt<len) {
				dp[cnt] = getdata();
				if (((ev&0xf0)==0x90 || (ev&0xf0)==0x80) && cnt==0) 
					UDR = 127-dp[cnt];
				else
					UDR = dp[cnt];
				cnt++;
			}
			if ((ev&0xf0) == 0x90)
				PORTB = ~dp[0];
			if ((ev&0xf0) == 0x80)
				PORTB = ~0x00;
/*			send_data(ev);
			for (cnt=0; cnt<len; cnt++)
				send_data(dp[cnt]);
*/		}
		

/*		if (tx_rdp == tx_wrp) {
			tx_buf[tx_wrp++] = 'A';
			tx_wrp &= TX_BUF_SIZE-1;
			send_byte();
		}
*/	}
}


ISR(SIG_UART_RECV) {
	rx_buf[rx_wrp++] = UDR;
	rx_wrp &= RX_BUF_SIZE-1;
	if (rx_wrp == rx_rdp) {
	// Buffer overflow!
		PORTB |= 0x20;
	}
}

ISR(SIG_UART_DATA) {
	if (tx_rdp != tx_wrp) {
		UDR = tx_buf[tx_rdp++];
		tx_rdp &= TX_BUF_SIZE-1;
	}
}
